from .lxc import (LXCManager, create_lxc, create_lxc_with_overlays,
        LXCWithOverlays, LXC, UnmanagedLXC, LXCLoader)
from .service import LXCService


def default_setup():
    """The default API setup for lxc4u

    This is the API that you access globally from lxc4u.
    """
    service = LXCService
    lxc_types = dict(LXC=LXC, LXCWithOverlays=LXCWithOverlays,
            __default__=UnmanagedLXC)
    loader = LXCLoader(lxc_types, service)
    manager = LXCManager(loader, service)
    return LXCAPI(manager=manager, service=service)


class LXCAPI(object):
    def __init__(self, manager=None, service=None):
        self._service = service
        self._manager = manager

    def create(self, name, base=None, overlays=None):
        if base or overlays:
            return create_lxc_with_overlays(name,
                    base, overlays, service=self._service)
        return create_lxc(name, service=self._service)

    def start(self, name):
        lxc = self._manager.get(name)
        lxc.start()
        return lxc

    def stop(self, name):
        lxc = self._manager.get(name)
        lxc.stop()
        return lxc

    def get(self, name):
        return self._manager.get(name)

    def list(self):
        return self._manager.list()



"""
lxc4u.service
~~~~~~~~~~~~~

This contains the low level access layer to any of the lxc command line tools.
"""
import os
import subwrap


def split_info_line(line):
    return map(str.strip, line.split(':'))


class LXCService(object):
    """Low level access layer to the lxc tools"""
    @classmethod
    def list_names(cls):
        """Lists all known LXC names"""
        response = subwrap.run(['lxc-ls'])
        output = response.std_out
        return map(str.strip, output.splitlines())

    @classmethod
    def lxc_path(cls, *join_paths):
        """Returns the LXC path (default on ubuntu is /var/lib/lxc)"""
        response = subwrap.run(['lxc-ls', '-d'])
        output = response.std_out
        lxc_path = output.splitlines()[0]
        lxc_path = lxc_path.strip()
        return os.path.join(lxc_path, *join_paths)

    @classmethod
    def create(cls, name, template=None):
        """Creates an LXC"""
        command = ['lxc-create', '-n', name]
        if template:
            command.extend(['-t', template])
        subwrap.run(command)

    @classmethod
    def start(cls, name):
        """Starts an LXC as a daemon

        This cannot start an LXC as a non-daemon. That doesn't make sense.
        """
        command = ['lxc-start', '-n', name, '-d']
        subwrap.run(command)

    @classmethod
    def stop(cls, name):
        """Stops a running LXC"""
        command = ['lxc-stop', '-n', name]
        subwrap.run(command)

    @classmethod
    def destroy(cls, name):
        command = ['lxc-destroy', '-n', name]
        subwrap.run(command)

    @classmethod
    def info(cls, name, get_state=True, get_pid=True):
        """Retrieves and parses info about an LXC"""
        # Run lxc-info quietly
        command = ['lxc-info', '-n', name]
        response = subwrap.run(command)
        lines = map(split_info_line, response.std_out.splitlines())
        return dict(lines)



import os
import json
from . import constants


class LXCMeta(object):
    """A dictionary like object that stores the metadata for an LXC"""

    @classmethod
    def load_from_file(cls, file_path):
        """Load the meta data given a file_path or empty meta data"""
        data = None
        if os.path.exists(file_path):
            metadata_file = open(file_path)
            data = json.loads(metadata_file.read())
        return cls(initial=data)

    def __init__(self, initial=None):
        initial = initial or {}
        self._metadata = initial.copy()

    def __getitem__(self, key):
        return self._metadata[key]

    def __setitem__(self, key, value):
        self._metadata[key] = value

    def get(self, key, default=None):
        return self._metadata.get(key, default)

    def set(self, key, value):
        self._metadata[key] = value

    def as_dict(self):
        return self._metadata.copy()

    def bind(self, lxc):
        """Bind to an LXC"""
        return BoundLXCMeta.bind_to_lxc(lxc, self)

    def bind_and_save(self, lxc):
        """Binds metadata to an LXC and saves it"""
        bound_meta = self.bind(lxc)
        bound_meta.save()
        return bound_meta


class BoundLXCMeta(object):
    """An LXCMeta wrapper that binds it to an LXC"""

    @classmethod
    def bind_to_lxc(cls, lxc, meta):
        return cls(lxc, meta)

    def __init__(self, lxc, meta):
        self._lxc = lxc
        self._meta = meta

    def __getitem__(self, key):
        return self._meta[key]

    def __setitem__(self, key, value):
        self._meta[key] = value

    def save(self):
        meta_file_path = self._lxc.path(constants.LXC_META_FILENAME)
        meta_file = open(meta_file_path, 'w')

        json_string = json.dumps(self._meta.as_dict())

        meta_file.write(json_string)
        meta_file.close()

    def as_dict(self):
        return self._meta.as_dict()
        return self._meta.as_dict()



import subwrap
from lxc import LXC

class LXCService(object):
    @classmethod
    def list_names(cls):
        """Lists all known LXC names"""
        response = subwrap.run(['lxc-ls'])
        output = response.std_out
        return map(str.strip, output.splitlines())

    def create(cls, *args, **kwargs):
        """Creates an LXC"""
        container = LXC.create(*args, **kwargs)
        return container



import tempfile
import overlay4u


class OverlayGroup(object):
    @classmethod
    def create(cls, end_dir, start_dir, overlays, overlay_temp_path=None):
        if not overlays:
            raise TypeError('overlays may not be an empty list')
        overlay_objects = []
        # Prime the loop
        current_lower = start_dir
        # Loop through all but the final overlay
        for overlay in overlays[:-1]:
            # Create a temporary directory to handle the mount points for any
            # intermediate overlays
            temp_mount_point = tempfile.mkdtemp(dir=overlay_temp_path)
            # Mount that overlay
            overlay_obj = overlay4u.mount(temp_mount_point,
                    current_lower, overlay)
            overlay_objects.append(overlay_obj)
            # The new lower directory should be the temporary
            # directory
            current_lower = temp_mount_point
        # Get the final overlay location
        overlay = overlays[-1]
        # Do the final mount point on the lxc_path using the name provided
        overlay_obj = overlay4u.mount(end_dir, current_lower, overlay)
        overlay_objects.append(overlay_obj)
        return cls(end_dir, start_dir, overlay_objects)

    @classmethod
    def from_meta(cls, meta):
        end_dir = meta[0]
        start_dir = meta[1]
        overlay_mount_points = meta[2]
        overlay_objects = map(lambda a: overlay4u.get(a), overlay_mount_points)
        return cls(end_dir, start_dir, overlay_objects)

    def __init__(self, end_dir, start_dir, overlays):
        self.end_dir = end_dir
        self.start_dir = start_dir
        self.overlays = overlays

    def unmount(self):
        for overlay in self.overlays:
            overlay.unmount()

    def meta(self):
        """Data for loading later"""
        mount_points = []
        for overlay in self.overlays:
            mount_points.append(overlay.mount_point)
        return [self.end_dir, self.start_dir, mount_points]

    def __iter__(self):
        return iter(self.overlays)

    def __len__(self):
        return len(self.overlays)



import os
import shutil
from .service import LXCService
from .overlayutils import OverlayGroup
from . import constants
from .meta import LXCMeta


class LXCAlreadyStarted(Exception):
    def __init__(self, name):
        message = 'LXC named "%s" is already running' % name
        super(LXCAlreadyStarted, self).__init__(message)


class LXCDoesNotExist(Exception):
    def __init__(self, name):
        message = 'LXC named "%s" does not exist' % name
        super(LXCDoesNotExist, self).__init__(message)


class LXCHasNoMeta(Exception):
    def __init__(self, name):
        message = 'LXC named "%s" has no BoundLXCMeta' % name
        super(LXCHasNoMeta, self).__init__(message)


class UnknownLXCType(Exception):
    pass


def create_lxc(name, template='ubuntu', service=None):
    """Factory method for the generic LXC"""
    service = service or LXCService
    service.create(name, template=template)
    meta = LXCMeta(initial=dict(type='LXC'))
    lxc = LXC.with_meta(name, service, meta, save=True)
    return lxc


class LXC(object):
    """The standard LXC"""
    @classmethod
    def with_meta(cls, name, service, meta, save=False):
        lxc = cls(name, service)
        lxc.bind_meta(meta, save=save)
        return lxc

    @classmethod
    def from_meta(cls, name, service, meta):
        return cls.with_meta(name, service, meta)

    def __init__(self, name, service):
        self.name = name
        self._service = service
        self._meta = None

    def start(self):
        """Start this LXC"""
        if self.status == 'RUNNING':
            raise LXCAlreadyStarted(self.name)
        self._service.start(self.name)

    def stop(self):
        """Stop this LXC"""
        self._service.stop(self.name)

    def destroy(self):
        self._service.destroy(self.name)

    @property
    def status(self):
        info = self._service.info(self.name)
        return info['state']

    @property
    def pid(self):
        info = self._service.info(self.name)
        return int(info['pid'])

    def path(self, *join_paths):
        return self._service.lxc_path(self.name, *join_paths)

    def __repr__(self):
        return '<%s "%s">' % (self.__class__.__name__, self.name)

    @property
    def meta(self):
        meta = self._meta
        if not meta:
            raise LXCHasNoMeta(self.name)
        return meta

    def bind_meta(self, meta, save=False):
        if save:
            self._meta = meta.bind_and_save(self)
        else:
            self._meta = meta.bind(self)


def create_lxc_with_overlays(name, base, overlays, overlay_temp_path=None,
        service=None):
    """Creates an LXC using overlays.

    This is a fast process in comparison to LXC.create because it does not
    involve any real copying of data.
    """
    service = service or LXCService
    # Check that overlays has content
    if not overlays:
        raise TypeError("Argument 'overlays' must have at least one item")

    # Get the system's LXC path
    lxc_path = service.lxc_path()
    # Calculate base LXC's path
    base_path = os.path.join(lxc_path, base)
    # Calculate the new LXC's path
    new_path = os.path.join(lxc_path, name)

    # Create the new directory if it doesn't exist
    if not os.path.exists(new_path):
        os.mkdir(new_path)

    overlay_group = OverlayGroup.create(new_path, base_path, overlays)
    initial_meta = dict(type='LXCWithOverlays',
            overlay_group=overlay_group.meta())
    meta = LXCMeta(initial=initial_meta)
    return LXCWithOverlays.with_meta(name, service, meta, overlay_group,
            save=True)


class UnmanagedLXCError(Exception):
    pass


class UnmanagedLXC(LXC):
    """An LXC that was created without lxc4u metadata"""
    def destroy(self, force=False):
        """UnmanagedLXC Destructor.

        It requires force to be true in order to work. Otherwise it throws an
        error.
        """
        if force:
            super(UnmanagedLXC, self).destroy()
        else:
            raise UnmanagedLXCError('Destroying an unmanaged LXC might not '
                'work. To continue please call this method with force=True')


class LXCWithOverlays(LXC):
    """An LXC that has an overlay group"""
    @classmethod
    def with_meta(cls, name, service, meta, overlay_group, save=False):
        lxc = cls(name, service, overlay_group)
        lxc.bind_meta(meta, save=save)
        return lxc

    @classmethod
    def from_meta(cls, name, service, meta):
        overlay_group_meta = meta.get('overlay_group')
        overlay_group = OverlayGroup.from_meta(overlay_group_meta)
        return cls.with_meta(name, service, meta, overlay_group)

    def __init__(self, name, service, overlay_group):
        self._overlay_group = overlay_group
        super(LXCWithOverlays, self).__init__(name, service)

    def destroy(self):
        """Unmounts overlay and deletes it's own directory"""
        self._overlay_group.unmount()
        shutil.rmtree(self.path())


class LXCLoader(object):
    def __init__(self, types, service):
        self._types = types
        self._service = service

    def load(self, name, meta):
        lxc_type_name = meta.get('type', constants.DEFAULT_LXC_TYPE_KEY)

        lxc_type_cls = self._types.get(lxc_type_name)
        if not lxc_type_cls:
            raise UnknownLXCType('LXC type "%s" is unknown' % lxc_type_name)
        return lxc_type_cls.from_meta(name, self._service, meta)


class LXCManager(object):
    """Manages the currently available LXCs"""
    def __init__(self, loader, service):
        self._service = service
        self._loader = loader

    def list(self):
        """Get's all of the LXC's and creates objects for them"""
        service = self._service

        lxc_names = service.list_names()
        lxc_list = []
        for name in lxc_names:
            lxc = self.get(name)
            lxc_list.append(lxc)
        return lxc_list

    def get(self, name):
        """Retrieves a single LXC by name"""
        lxc_meta_path = self._service.lxc_path(name,
                constants.LXC_META_FILENAME)
        meta = LXCMeta.load_from_file(lxc_meta_path)
        lxc = self._loader.load(name, meta)
        return lxc



from .api import default_setup

__all__ = ['create', 'start', 'stop', 'get', 'list']

__default__ = default_setup()

create = __default__.create
start = __default__.start
stop = __default__.stop
get = __default__.get
list = __default__.list
